;;; ambsimple.scm - Clement JONQUET - PhD student in Compuer Science - LIRMM - September 2003 ;;;

; This module load a non deterministic interpreter based on simple.scm


;;; ambevaluate-... functions ;;;

(define amb? (test-tag 'amb))

(define (amb-choices exp) (cdr exp))

(define ambevaluate
  (lambda (e r ks kf)
;    (display "Evaluate de : ")(display e)(newline)
    ((if (constant? e) 
	 ambevaluate-constant 
     (if (variable? e) 
	 ambevaluate-variable 
     (if (quasiquote? e)
	 ambevaluate-quasiquote
     (if (if? e)
	 ambevaluate-if
     (if (assignment? e) 
	 ambevaluate-assignment
     (if (define? e)
	 ambevaluate-define
     (if (let? e)
	 ambevaluate-let
     (if (begin? e)
	 ambevaluate-begin
     (if (amb? e)
	 ambevaluate-amb
     (if (abstraction? e) 
	 ambevaluate-abstraction
	 ambevaluate-combination))))))))))
     e r ks kf)))

(define  ambevaluate-constant
  (lambda (e r ks kf) 
    (ks (constant-part e) kf)))

(define  ambevaluate-variable
  (lambda (e r ks kf) 
    (get-pair e r 
	      (lambda (success-pair) (ks (cdr success-pair) kf))
	      (lambda () (wrong "symbol not bound: " e)))))

(define ambevaluate-quasiquote
  (lambda (e r ks kf)
    (define (quasifoo exp) 
      (if (null? exp) 
	  ()
	  (if (pair? exp)
	      (if (unquote? (car exp))
		  (ambevaluate (car (cdr (car exp)))
			       r
			       (lambda (value kf) (cons value (quasifoo (cdr exp))))
			       kf)
		  (if (unquote-splicing? (car exp))
		      ; Be carefull...it's not good
		      (ambevaluate (car (cdr (car exp)))
				   r
				   (lambda (value kf) (append value (quasifoo (cdr exp))))
				   kf)
		      (cons (quasifoo (car exp)) (quasifoo (cdr exp)))))
	      exp)))
    (ks (quasifoo (car (cdr e))) kf)))

(define  ambevaluate-if 
  (lambda (e r ks kf) 
    (ambevaluate (test-part e)
		 r 
		 (lambda (value kf2) 
		   (if value (ambevaluate (then-part e) r ks kf2) (ambevaluate (else-part e) r ks kf2)))
		 kf)))

(define ambevaluate-assignment
  (lambda (e r ks kf) 
    (ambevaluate (value-part e)
		 r 
		 (lambda (value kf2) 
		   (let ((old-value value))
		     (get-pair (id-part e)
			       r 
			       (lambda (success-pair)
				 (set-cdr! success-pair value)
				 (ks (void) (lambda () (set-cdr! success-pair old-value) (kf2))))
			       (lambda () 
				 (set-cdr! r (cons (cons (id-part e) value) (cdr r)))
				 (ks (void) (lambda () (set-cdr! r old-value) (kf2)))))))
		 kf)))

(define ambevaluate-define
  (lambda (e r ks kf)
    (if (pair? (car (cdr e)))
        ; for call like (define (namefunc param) (body)) 
	(ambevaluate (append (list 'lambda (cdr (car (cdr e)))) (cdr (cdr e))) 
;	(ambevaluate (list 'lambda (cdr (car (cdr e))) (value-part e)) ;build (lambda (param) (body))
		     r 
		     (lambda (value kf2) 
		       (let ((old-value value))
			 (get-pair (car (car (cdr e))) r 
				   ; If the symbol already exists is value is changed
				   (lambda (success-pair)
				     (set-cdr! success-pair value)
				     (ks (void) (lambda () (set-cdr! success-pair old-value) (kf2))))
				   ; Else a new bindings is added to r
				   (lambda () 
				     (set-cdr! r (cons (cons (car (car (cdr e))) value) (cdr r)))
				     (ks (void) (lambda () (set-cdr! r old-value) (kf2)))))))
		     kf)
	; for a call like (define var val) or (define func (lambda param body))
	(ambevaluate (value-part e)
		     r
		     (lambda (value kf2) 
		       (let ((old-value value))
			 (get-pair (id-part e) r 
				 ; If the symbol already exists is value is changed
				 (lambda (success-pair)
				   (set-cdr! success-pair value)
				   (ks (void) (lambda () (set-cdr! success-pair old-value) (kf2))))
				 ; Else a new bindings is added to r
				 (lambda () 
				   (set-cdr! r (cons (cons (id-part e) value) (cdr r)))
				   (ks (void) (lambda () (set-cdr! r old-value) (kf2)))))))
		     kf))))

(define ambevaluate-let
  (lambda (e r ks kf)
    (ambevaluate (append (list (append (list 'lambda (let-var e)) (let-body e))) (amblet-init e)) r ks kf)))
;    (ambevaluate-sequence (let-body e) (extend r (let-var e) (amblet-init e r)) ks kf)))

;(define amblet-init
;  (lambda (e r)
;    (mapper (lambda (x) (ambevaluate (car (cdr x)) r ambreturn ambnothing)) (car (cdr e)))))

(define amblet-init
  (lambda (e) (mapper (lambda (x) (car (cdr x))) (car (cdr e)))))

; ambevaluate-sequence marche pour l'xp mais chie pour l'eval non deterministe...

(define ambevaluate-begin
  (lambda (e r ks kf)
    (ambevaluate-sequence (cdr e) r ks kf)))

(define ambevaluate-sequence
  (lambda (body r ks kf)
    (if (null? (cdr body)) 
	(ambevaluate (car body) r ks kf) 
	(ambevaluate (car body) 
		     r
		     (lambda (value kf2) (ambevaluate-sequence (cdr body) r ks kf2))
		     kf))))

(define (ambevaluate-amb e r ks kf)
  (define (try-next choices)
    (if (null? choices)
	(kf)
	(ambevaluate (car choices) r ks (lambda () (try-next (cdr choices))))))
    (try-next (amb-choices e)))

(define ambevaluate-abstraction
  (lambda (e r ks kf)
    (ks (make-compound (formals-part e) (body-part e) r) kf)))

(define ambevaluate-combination
  (lambda (e r ks kf) 
    (ambevaluate (operator-part e)
	      r
	      (lambda (proc kf2) 
		(if (reifier? proc)
		    ; reifying procedures non gres...
		    ((reifier-to-compound proc) (operands-part e) r k) 
		    (ambevaluate-operands (operands-part e)
					  r 
					  (lambda (args kf3) (ambapply-procedure proc args ks kf3))
					  kf2)))
	      kf)))

(define ambevaluate-operands
  (lambda (operands r ks kf) 
    (if (null? operands)
	(ks '() kf)
	(ambevaluate (car operands)
		     r 
		     (lambda (arg kf2) (ambevaluate-operands (cdr operands)
							     r
							     (lambda (args kf3) (ks (cons arg args) kf3))
							     kf2))
		     kf))))

;;; ambapply-procedure function ;;;

(define ambapply-procedure
  (lambda (proc args ks kf) 
    (if (compound? proc)
	(ambevaluate-sequence (procedure-body proc)
			      (extend (procedure-environment proc) (procedure-parameters proc) args)
			      ks
			      kf)
	(ks (apply-primitive (procedure-name proc) args) kf))))

;;; require procedure ;;;

(define require (lambda (p) (if (not p) (amb) (void))))

(define ambreturn (lambda (val kf) val))
(define ambnothing (lambda () ()))

;;; Read-Eval-Print loop and how load a file ;;;

(define ambopenloop
  (lambda (read-prompt write-prompt)
    (define (internal-loop try-again)
      (newline)(display read-prompt)
      (let ((input (read)))
	(if (eq? input 'try-again)
	    (try-again)
	    (begin (newline)(display "Starting a new problem..")
		   (ambevaluate input
				global-env
				(lambda (val next-alternative)
				  (newline)(display write-prompt)
				  (if (eq? val (void))
				      "Nothing will be displayed" 
				      (write val))
				  (newline)(internal-loop next-alternative))
				(lambda () 
				  (newline) (display "No more values...")
				  (ambopenloop read-prompt write-prompt)))))))
    (internal-loop
     (lambda () (newline)(display "The is no current problem...")(ambopenloop read-prompt write-prompt)))))
     
(define ambloadfile
  (lambda (file)
    ((lambda (port) 
       ((lambda (loop) 
	  (set! loop 
		(lambda (v) 
		  (if (eof-object? v)
		      (close-input-port port)
		      (ambevaluate v 
				   global-env
				   (lambda (ignore kignore) (loop (read port)))
				   (lambda () (display "Erreur lors du chargement..."))))))
	  (loop (read port)))
	'*))
     (open-input-file file))))

;;; The bootin process ;;;

(define ambboot-flat
  (lambda () 
    (initialize-global-env) (ambopenloop "AMB: 0< " "AMB: 0> ")))

(define ambboot-tower
  (lambda () 
    (initialize-global-env)
    (ambloadfile "simple.scm")
    (ambloadfile "ambsimple.scm")
    (set-cdr! global-env (cons (cons 'global-env global-env) (cdr global-env)))
    (ambopenloop "AMB: 0> " "AMB: 0: ")))


