;;; evalteacher.scm - Clement JONQUET - PhD student in Compuer Science - LIRMM - September 2003 ;;;

; This file gives the specific interpreter for the teacher, it allows the interpreter described in simple.scm to interpret kqmlmsg.

;;; kqmlmsg process ;;;

(define kqmlmsg? (test-tag 'kqmlmsg))

(define (make-kqmlmsg p s r c)
  (list 'kqmlmsg p s r c))
 
(define (performative kqmlmsg) (car (cdr kqmlmsg)))
(define (sender kqmlmsg) (car (cdr (cdr kqmlmsg))))
(define (receiver kqmlmsg) (car (cdr (cdr (cdr kqmlmsg)))))
(define (content kqmlmsg) (car (cdr (cdr (cdr (cdr kqmlmsg))))))

;;; evaluate procedure ;;;

(set! ambevaluate
  (lambda (e r ks kf)
;    (newline)(display "Evaluate de : ")(display e)(newline)
    ((if (constant? e) 
	 ambevaluate-constant 
     (if (variable? e) 
	 ambevaluate-variable 
     (if (quasiquote? e)
	 ambevaluate-quasiquote
     (if (if? e)
	 ambevaluate-if
     (if (assignment? e) 
	 ambevaluate-assignment
     (if (define? e)
	 ambevaluate-define
     (if (let? e)
	 ambevaluate-let
     (if (begin? e)
	 ambevaluate-begin
     (if (amb? e)
	 ambevaluate-amb
     (if (kqmlmsg? e)
	 ambevaluate-kqmlmsg-and-broadcast
     (if (abstraction? e) 
	 ambevaluate-abstraction
	 ambevaluate-combination)))))))))))
     e r ks kf)))

; Process kqmlmsg, includes broacast performative, return a list of messages or a data...

(define (ambevaluate-kqmlmsg-and-broadcast e r ks kf)
  (let ((performative (performative e))
	(sender (sender e))
	(receiver (receiver e))
	(content (content e))
	(interpreter ambevaluate))

    (if (eq? performative 'assertion)
	(interpreter content r 
		     (lambda (val kf2) (ks (list (make-kqmlmsg 'ack receiver sender val)) kf2))
		     kf)
    (if (eq? performative 'request)
	(interpreter content r 
		     (lambda (val kf2) (ks (list (make-kqmlmsg 'answer receiver sender val)) kf2))
		     kf)
    (if (eq? performative 'order)
	(interpreter content r 
		     (lambda (val kf2) (ks (list (make-kqmlmsg 'executed receiver sender val)) kf2))
		     kf)
    (if (eq? performative 'ack)	'ok
    (if (eq? performative 'answer) (ks (analyze-answer sender receiver content) kf)
    (if (eq? performative 'executed) (ks (analyze-answer sender receiver content) kf)
   ;BROADCAST : content as (perform cont)
    (if (eq? performative 'broadcast)
	(ks (mapper (lambda (x) (make-kqmlmsg (car content) receiver (car x) (car (cdr content))))
		    current-conversations)
	    kf)
	;Else
	(ks (list (make-kqmlmsg 'answer receiver sender `('no-such-performative ,performative))) kf))))))))))

; performative-to-learn is a procedure which process some performative to learn... it returns a list of message...empty by default

(define (analyze-answer sen rec con)
  (if (list? con)
      (if (eq? (car con) 'no-such-performative)
	  ;Call performative-to-learn function
	  (performative-to-learn sen rec (car (cdr con)))
	  ;Else
	  (begin (display "Unknown performative ")(display (car (cdr con)))(display " and no method to teach it")
		 con))
      con))

(define (performative-to-learn sender receiver unknowperform)
  (if (eq? unknowperform 'broadcast)
      ;How to learn broadcast performative... 
      (let ((kqmlmsg1 (list 'kqmlmsg 'assertion receiver sender learn-broadcast-code-msg))
	    (kqmlmsg2 (list 'kqmlmsg 'assertion receiver sender learn-broadcast-msg)))
	(list kqmlmsg1 kqmlmsg2))
      ;Else... send a message sorry
      (list (list 'kqmlmsg 'answer receiver sender '(sorry)))))
      
(define learn-broadcast-code-msg
  '(define learn-broadcast-code 
     (let ((oldbody (car (cdr (cdr ambevaluate-kqmlmsg))))
	   (defenv (car (cdr (cdr (cdr ambevaluate-kqmlmsg))))))
       (list 'compound
	     '(e r ks kf)
	     (list (list (car (car oldbody)) 
			 (car (cdr (car oldbody)))
			 (list 'if
			       '(eq? performative 'broadcast)
			       '(ks (mapper (lambda (x) (make-kqmlmsg (car content) receiver (car x) (car (cdr content))))
					    current-conversations)
					   kf)
			       (car (cdr (cdr (car oldbody)))))))
	     defenv))))

(define learn-broadcast-msg 
  '(set! ambevaluate-kqmlmsg learn-broadcast-code))

; This is the square procedure...       
(define (square x) (* x x))  
