;;; simple.scm - Clement JONQUET - DEA Informatique LIRMM - Mai 2003 ;;;

; This file described a meta-evaluator based on Jefferson and Friedman's article : A Simple Reflective Interpreter

;;; evaluate Procedure and sub-procedures ;;;

(define  evaluate
  (lambda (e r k) 
;    (display "evaluate de : ")(display e)(newline)
         ((if (constant? e) 
	      evaluate-constant 
	 (if (variable? e) 
	     evaluate-variable 
	 (if (quasiquote? e)
	     evaluate-quasiquote
	 (if (if? e)
	     evaluate-if
	 (if (assignment? e) 
	     evaluate-assignment
	 (if (define? e)
	     evaluate-define
         (if (let? e)
	     evaluate-let
	 (if (begin? e)
	     evaluate-begin
	 (if (abstraction? e) 
	     evaluate-abstraction
	     evaluate-combination)))))))))
     e r k)))

(define  evaluate-constant
  (lambda (e r k) 
    (k (constant-part e))))

(define  evaluate-variable
  (lambda (e r k) 
    (get-pair e r 
	      (lambda (success-pair) (k (cdr success-pair)))
	      (lambda () (wrong "symbol not bound: " e)))))

(define evaluate-quasiquote
  (lambda (e r k)
    (define (quasifoo exp) 
      (if (null? exp) 
	  ()
	  (if (pair? exp)
	      (if (unquote? (car exp))
		  (cons (evaluate (car (cdr (car exp))) r return) (quasifoo (cdr exp)))
		  (if (unquote-splicing? (car exp))
		      ; Be carefull...it's not good
		      (append (evaluate (car (cdr (car exp))) r return) (quasifoo (cdr exp)))
		      (cons (quasifoo (car exp)) (quasifoo (cdr exp)))))
	      exp)))
    (k (quasifoo (car (cdr e))))))

(define  evaluate-if 
  (lambda (e r k) 
    (evaluate (test-part e) r 
	      (lambda (v) (if v (evaluate (then-part e) r k) (evaluate (else-part e) r k))))))

(define evaluate-assignment
  (lambda (e r k) 
    (evaluate (value-part e) r 
	      (lambda (v) 
		(get-pair (id-part e) r 
			  (lambda (success-pair)
			    (set-cdr! success-pair v) (k (void)))
			  (lambda () 
			    (set-cdr! r (cons (cons (id-part e) v) (cdr r)))
				  (k (void))))))))

(define evaluate-define
  (lambda (e r k)
    (if (pair? (car (cdr e)))
        ; for call like (define (namefunc param) (body)) 
	(evaluate (append (list 'lambda (cdr (car (cdr e)))) (cdr (cdr e))) r 
		  (lambda (v)
		    (get-pair (car (car (cdr e))) r 
			      ; If the symbol already exists is value is changed
			      (lambda (success-pair)
				(set-cdr! success-pair v) (k (void)))
			      ; Else a new bindings is added to r
			      (lambda () 
				(set-cdr! r (cons (cons (car (car (cdr e))) v) (cdr r)))
				(k (void))))))
	; for a call like (define var val) or (define func (lambda param body))
	(evaluate (value-part e) r
		  (lambda (v) 
		    (get-pair (id-part e) r 
			      ; If the symbol already exists is value is changed
			      (lambda (success-pair)
				(set-cdr! success-pair v) (k (void)))
			      ; Else a new bindings is added to r
			      (lambda () 
				(set-cdr! r (cons (cons (id-part e) v) (cdr r)))
				(k (void)))))))))

(define evaluate-let
  (lambda (e r k)
    (evaluate-sequence (let-body e) (extend r (let-var e) (let-init e r)) k)))

(define evaluate-begin
  (lambda (e r k)
    (evaluate-sequence (cdr e) r k)))

(define evaluate-abstraction
  (lambda (e r k)
    (k (make-compound (formals-part e) (body-part e) r))))

(define evaluate-combination
  (lambda (e r k) 
    (evaluate (operator-part e)
	      r
	      (lambda (proc) 
		(if (reifier? proc)
		    ((reifier-to-compound proc) (operands-part e) r k) 
		    (evaluate-operands (operands-part e) r 
				       (lambda (args) (apply-procedure proc args k))))))))

(define evaluate-operands
  (lambda (operands r k)
    (if (null? operands)
	(k '())
	(evaluate (car operands) r 
		  (lambda (v) (evaluate-operands (cdr operands) r (lambda (w) (k (cons v w)))))))))

(define evaluate-sequence
  (lambda (body r k)
    (if (null? (cdr body)) 
	(evaluate (car body) r k) 
	(evaluate (car body) r (lambda (v) (evaluate-sequence (cdr body) r k))))))

;;; apply-procedure and apply-primitive functions ;;;

(define apply-procedure
  (lambda (proc args k) 
    (if (compound? proc)
	(evaluate-sequence (procedure-body proc)
			   (extend (procedure-environment proc) (procedure-parameters proc) args)
			   k)
	(k (apply-primitive (procedure-name proc) args)))))

(define apply-primitive
  (lambda (name args)
        (if (eq? name 'car)
	    (car (1st args))
	(if (eq? name 'cdr)
	    (cdr (1st args))
	(if (eq? name 'cons)
	    (cons (1st args) (2nd args))
	(if (eq? name 'set-car!)
	    (set-car! (1st args) (2nd args))
	(if (eq? name 'set-cdr!)
	    (set-cdr! (1st args) (2nd args))
	(if (eq? name 'assq)
	    (assq (1st args) (2nd args)) 
	(if (eq? name 'memq) 
	    (memq (1st args) (2nd args))
	(if (eq? name 'null?)
	    (null? (1st args))
	(if (eq? name '=)
	    (= (1st args) (2nd args)) 
	(if (eq? name 'eq?)
	    (eq? (1st args) (2nd args))
	(if (eq? name 'equal?)
	    (equal? (1st args) (2nd args))
	(if (eq? name 'newline) 
	    (newline)
	(if (eq? name 'write)
	    (write (1st args)) 
	(if (eq? name 'display)
	    (display (1st args))
	(if (eq? name 'read)
	    (if (null? args) (read) (read (1st args)))
	(if (eq? name '+)
	    (+ (1st args) (2nd args))
	(if (eq? name '-)
	    (- (1st args) (2nd args))
	(if (eq? name '*)
	    (* (1st args) (2nd args))
	(if (eq? name '/)
	    (/ (1st args) (2nd args))
	(if (eq? name '<)
	    (< (1st args) (2nd args))
	(if (eq? name '>)
	    (> (1st args) (2nd args))
	(if (eq? name '<=)
	    (<= (1st args) (2nd args))
	(if (eq? name '>=)
	    (>= (1st args) (2nd args))
	(if (eq? name 'symbol?)
	    (symbol? (1st args))
	(if (eq? name 'list)
	    args
	(if (eq? name 'append)
	    (append (1st args) (2nd args))
	(if (eq? name 'pair?)
	    (pair? (1st args))
	(if (eq? name 'list?)
	    (list? (1st args))
	(if (eq? name 'length)
	    (length (1st args))
	(if (eq? name 'reverse)
	    (reverse (1st args))
	(if (eq? name 'procedure?)
	    (procedure? (1st args))
	(if (eq? name 'eof-object?)
	    (eof-object? (1st args))
	(if (eq? name 'close-input-port) 
	    (close-input-port (1st args))
	(if (eq? name 'open-input-file)
	    (open-input-file (1st args))
	(if (eq? name 'void) 
	    (void)
	(if (eq? name 'not)
	    (not (1st args))
	(if (eq? name 'abs)
	    (abs (1st args))
	(if (eq? name 'member)
	    (member (1st args) (2nd args))
	    "Shouldn't Happen!"
	    ))))))))))))))))))))))))))))))))))))))))

;;; Environment's traitement ;;;

(define extend
  (lambda (r ids vals)
    (if (null? ids)
	r
	(extend (cons (cons (car ids) (car vals)) r) (cdr ids) (cdr vals)))))

(define find-pair
  (lambda (elt alist success failure) 
    ((lambda (assq-result) 
       (if (null? assq-result)
	   (failure)
	   (success assq-result)))
     (assq elt alist))))

; get-pair return an item of an environment (name type param body defenv)
; If an item is not found in the given environment, it is not look for in the global-env
(define get-pair
  (lambda (id r success failure) 
    ;(find-pair id r success (lambda () (find-pair id global-env success failure)))))
    (find-pair id r success failure)))


(define empty-env (list (cons 'zeo 'zeo)))

(define global-env (list (cons 'zeo 'zeo)))

(define (before item alist)
  (if (pair? alist)
      (if (eq? (car alist) 'globalEnv)
	  ()
	  (if (eq? item (car (car alist)))
	      ()
	      (cons (car alist) (before item (cdr alist)))))
      ()))

(define copy-env
  (lambda (env)
    (let ((newenv (list (cons 'aze 'aze))))
      (define env-change
	(lambda (x) 
	  (if (list? x)
	      (if (eq? 'compound (2nd x))
		  (list (1st x) 'compound (3rd x) (4th x) (append (before 'member (5th x)) newenv))
		  x)
	      x)))
      (let ((goodenv (mapper env-change env)))
	(set-cdr! newenv goodenv)
	goodenv))))

(define primitive-identifiers
  (lambda () 
    '(car cdr cons set-car! set-cdr! assq memq null? = eq? equal? newline write display read + - * / < > <= >= symbol? list append pair? list? length reverse eof-object? close-input-port open-input-file void not abs member)))

(define initialize-global-env
  (lambda ()
    (set! global-env 
	  (extend empty-env (primitive-identifiers) (mapper make-primitive (primitive-identifiers))))
    (void))) 

;;; List utilities and other usefull procedures ;;; 

(define 1st (lambda (l) (car l)))
(define 2nd (lambda (l) (car (cdr l))))
(define 3rd (lambda (l) (car (cdr (cdr l)))))
(define 4th (lambda (l) (car (cdr (cdr (cdr l))))))
(define 5th (lambda (l) (car (cdr (cdr (cdr (cdr l)))))))
(define test-tag (lambda (tago) (lambda (e) (if (pair? e) (eq? (car e) tago) #f))))

(define void
  (lambda () (cons '* '*))) 

(define  wrong 
  (lambda (message object)
    (display "Error: ") (display message) (display object) (newline)))

(define mapper
  (lambda (f l) 
    (if (null? l)
	'()
	(cons (f (car l)) (mapper f (cdr l))))))

(define return (lambda (x) x))
(define nothing (lambda () ()))

;;; Procedures use by the evaluating procedures ;;; 

(define let-body 
  (lambda (exp) (cdr (cdr exp))))

(define let-var
  (lambda (exp) (mapper (lambda (x) (car x)) (car (cdr exp)))))

(define let-init
  (lambda (e r)
    (mapper (lambda (x) (evaluate (car (cdr x)) r return)) (car (cdr e)))))

(define compound-to-reifier
  (lambda (compound) (cons 'reifier (cdr compound))))

(define reifier-to-compound
  (lambda (reifier) (cons 'compound (cdr reifier))))

(define make-compound
  (lambda (formals body r) (list 'compound formals body r)))

(define make-primitive
  (lambda (op) (list 'primitive op)))

(define make-reifier
  (lambda (formals body r) (list 'reifier formals body r)))

(define procedure-parameters 2nd)

(define procedure-body 3rd)

(define procedure-environment 4th)

(define procedure-name 2nd)

(define constant-part
  (lambda (e) (if (quote? e) (2nd e) e))) 

(define test-part 2nd) 

(define then-part 3rd) 

(define else-part 4th) 

(define id-part 2nd)

(define value-part 3rd) 

(define formals-part 2nd) 

(define body-part 
  (lambda (e) (cdr (cdr e))))

(define operator-part 1st) 

(define operands-part cdr)

;;; Syntax's procedures ;;;

(define compound? (test-tag 'compound)) 

(define primitive? (test-tag 'primitive))

(define reifier? (test-tag 'reifier))

(define variable?
  (lambda (e) 
    (if ((test-tag 'define) e) #f (symbol? e))))

(define if? (test-tag 'if)) 

(define assignment? (test-tag 'set!)) 

(define define? (test-tag 'define))

(define let? (test-tag 'let))

(define begin? (test-tag 'begin))

(define abstraction? (test-tag 'lambda)) 

(define constant? 
  (lambda (e) 
    (if (pair? e) (quote? e) (if (symbol? e) #f #t)))) 

(define quote? (test-tag 'quote)) 

(define quasiquote? (test-tag 'quasiquote))

(define unquote? (test-tag 'unquote))

(define unquote-splicing? (test-tag 'unquote-splicing))

;;; Read-Eval-Print loop and how load a file ;;;

(define openloop
  (lambda (read-prompt write-prompt)
    (newline)(display read-prompt)
    (evaluate (read) global-env (lambda (v) 
				  (newline)(display write-prompt)
				  (if (eq? v (void))
				      "Nothing will be displayed" 
				      (write v)) (newline) (openloop read-prompt write-prompt)))))
(define loadfile
  (lambda (file)
    ((lambda (port) 
       ((lambda (loop) 
	  (set! loop 
		(lambda (v) 
		  (if (eof-object? v)
		      (close-input-port port) 
		      (evaluate v global-env (lambda (ignore) (loop (read port)))))))
	  (loop (read port)))
	'*))
     (open-input-file file))))

;;; The booting process ;;;

(define boot-flat
  (lambda () 
    (initialize-global-env) (openloop "0< " "0> ")))

(define boot-tower
  (lambda () 
    (initialize-global-env) 
    (loadfile "simple.scm")
    (set-cdr! global-env (cons (cons 'global-env global-env) (cdr global-env)))
    (openloop "0> " "0: "))) 

(define my-boot-tower 
  (lambda () 
    (initialize-global-env) 
    (loadfile "simple.scm")
    (set-cdr! global-env (cons (cons 'global-env global-env) (cdr global-env))))) 

;;; Printing procedures ;;;

(define print-env
  (lambda (env)
    (mapper 
     (lambda (x) 
       (if (list? x)
	   ; Si c'est un list
	   (begin (display "-- ")(display (1st x))(display " : ")
		  (if (eq? 'global-env (1st x))
		      (display "...")
		      (if (eq? 'compound (2nd x))
			  (begin (display "compound ")(display (3rd x))(display " ")
				 (display (4th x))(display " ")
				 (display "..."))
;				 (newline)(display "DEFENV: ") (print-proc (3rd (5th x))))
		      (if (eq? 'primitive (2nd x))
			  (begin (display "primitive ")(display (3rd x)))
		      (if (eq? 'reifier (2nd x))
			  (begin (display "reifier ")(display (3rd x))(display " ")
				 (display (4th x))(display " ")(display "..."))
			  (display (cdr x))))))
		  (newline))
	   ; Sinon c'est une paire pointe
	   (begin (display "-- ")(display (car x))(display " : ")(display (cdr x))(newline))))
     env)
    (display "-----------------Environement Printed")))

(define print-proc
  (lambda (proc)
    (display (1st proc))(display " ")
    (display (2nd proc))(display " ")
    (if (compound? proc)
	(begin (display (3rd proc))(display " ")
	       (display "...")(newline))
	       ;(print-env (4th proc))(newline))
	(display "..."))))

(define proc-body
  (lambda (proc)
    (car (car (cdr (cdr proc))))))